#include <cstdio>
#include <cstring>
#include <cmath>

using namespace std;

#include "snd.hh"
#include "instrument.hh"
#include "sndplayer.hh"
#include "analysis.hh"
#include "s3m.hh"

static void DumpSND(const song &song)
{
    for(unsigned a=0; a<repository.size(); ++a)
    {
        fprintf(stderr, "Instrument %u:\n", a+1);
        DumpSample(repository[a]);
    }

    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const SNDChannel &tmp = song.ChanPtr(a, c);
            char Buf[64];
            int colormode=0;
            
            sprintf(Buf, "v%Xs%dw%ua%u", tmp.vol, tmp.sample, tmp.wavelen, tmp.age);
            if(tmp.ins)
                sprintf(strchr(Buf,0), "(%u)", tmp.ins);
            
            sprintf(Buf+32, "%18s", Buf);
            if(tmp.vol)
            {
                if(IsBeat(a, c))
                {
                    fputs("\33[1;34m", stderr);
                    colormode=2;
                }
                else
                    colormode=1;
            }
            else
            {
                colormode=0;
                fputs("\33[1;30m", stderr);
            }
            for(char *s=Buf+32; *s; ++s)
            {
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[1;37m", stderr);
                */
                fputc(*s, stderr    );
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[m", stderr);
                */
            }
            fputs("\33[m", stderr);
        }
        fprintf(stderr, "\n");
    }
}


static unsigned NesVoltoS3Mvol(unsigned v)
{
    static unsigned char voltable[16] =
    {0,32,38,40,
     42,46,48,50,
     52,54,56,58,
     60,62,63,64};
    return voltable[v];
}

static void Patternize(const song &song)
{
    S3M musicfile;
    S3M::eventtable events;
    
    events.setgeometry(0, song.channelcount);
    
    /* Handle each channel separately */
    for(unsigned chan=0; chan<song.channelcount; ++chan)
    {
        vector<S3M::event> chanevents;
        
        switch(song.ChanPtr(0, chan).type)
        {
        	case SNDChannel::pcm:
        		musicfile.setupchannel(chan, S3M::leftchan(chan));
        		break;
        	default:
        		musicfile.setupchannel(chan, S3M::amechan(chan));
        		break;
        }
        
        unsigned lastrow=0;
        bool wason = false;
        for(double pos=0; ; pos += TickSpacer)
        {
            unsigned row = (unsigned)pos;
            if(row >= song.rownum)break;
        
            const SNDChannel &tmp = song.ChanPtr(row, chan);
            
            // vol=tmp.vol     (volume(0..15))
            // smp=tmp.sample  (sqr:duty; pcm:pcm sample no)
            // wav=tmp.wavelen (wave length)
            // age=tmp.age     (age since last touched)
            // ins=tmp.ins     (index to repository)
            // beat=beat
            const SNDChannel *beatrow = NULL;
            while(lastrow <= row)
                if(IsBeat(lastrow++, chan))
                	beatrow = &song.ChanPtr(lastrow-1, chan);
            
            S3M::event ev;
            
            if(beatrow)
            {
            	double div = 8000.0;
            	if(tmp.type == SNDChannel::rnd)
            	{
            		div = 250.0;
            	}
            	else if(tmp.type == SNDChannel::tri)
            	{
            		div = 1000.0;
            	}
                unsigned basewave = tmp.wavelen;
                unsigned nuotti = (unsigned)(12 * log(div / basewave) / log(2.0));
                unsigned oct = nuotti/12;
                ev.note = (nuotti%12) + oct*16;
                wason = true;
            }
            else if(wason && !tmp.vol)
            {
            	ev.note = 254;
            	ev.vol = 0;
                wason = false;
            }
            
            if(tmp.type == SNDChannel::pcm && beatrow && tmp.sample)
                ev.ins = tmp.sample + repository.size();
            else if(tmp.type == SNDChannel::pcm && beatrow)
            	ev.ins = beatrow->sample + repository.size();
            else if(tmp.ins)
            	ev.ins = tmp.ins;
            else if(beatrow && beatrow->ins)
            	ev.ins = beatrow->ins;

            if(beatrow)
            {
                int volavg = 0, volcount=0;
                for(unsigned tmprow = row; tmprow < row+8 && tmprow < song.rownum; ++tmprow)
                {
                    if(tmprow > row && IsBeat(tmprow, chan))break;
                    const SNDChannel &tmp = song.ChanPtr(tmprow, chan);
                    if(!tmp.vol)break;
                    volavg += tmp.vol; ++volcount;
                }
                volavg /= volcount;
                ev.vol = NesVoltoS3Mvol(volavg);
            }
            else
            {
            	//ev.vol = NesVoltoS3Mvol(tmp.vol);
            }
            
            chanevents.push_back(ev);
        }
        events.loadchannel(chan, chanevents);
    }

    fprintf(stderr, "----------\n");
    for(unsigned a=0; a<events.rowcount; ++a)
    {
        fprintf(stderr, "%u ", a % 4);
    	for(unsigned c=0; c<events.channelcount; ++c)
    	{
    		const S3M::event& ev = events.events[c + events.channelcount * a];
    		
            switch(ev.note)
            {
            	case 255: fprintf(stderr, "..."); break;
            	case 254: fprintf(stderr, "^^^"); break;
            	default:
		            char Buf[64];
        		    static const char nuotit[12*3+1] = "C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
		            sprintf(Buf, "%s%d", nuotit+3*(ev.note%16), ev.note/16);
        		    fprintf(stderr, "%s", Buf);
        	}
        	if(ev.ins) fprintf(stderr, " %02u", ev.ins);
        	else fprintf(stderr, " ..");
        	if(ev.vol) fprintf(stderr, " %02u", ev.vol);
        	else fprintf(stderr, " ..");
            
            fprintf(stderr, "  ");
    	}
        fprintf(stderr, "\n");
    }
    unsigned patcount = (events.rowcount + 64 - 1) / 64;
    for(unsigned a=0; a<patcount; ++a)
    {
    	S3M::eventtable tmp(events, a*64, 64);
    	musicfile.getpattern(a) = musicfile.compresspattern(tmp, 64);
    	musicfile.setorder(a, a);
    }
    musicfile.setorder(patcount, 255);
    musicfile.setorder(patcount+1, 255);
    
    for(unsigned a=0; a<256; ++a)
    	if(song.dpcmcache[a].size())
    	{
    		S3M::instrument &ins = musicfile.getinstrument(a + repository.size());
    		
    		ins.MakeSMP();
    		ins.audio = song.dpcmcache[a];
    		ins.SetLoop(0, ins.audio.size());
    	}
    for(unsigned a=0; a<repository.size(); ++a)
    {
    	S3M::instrument &ins = musicfile.getinstrument(a);
    	ins.type = 2;
    	
    }
    
    musicfile.Save("test.s3m");
}

int main(int argc, const char *const *argv)
{
    if(argc != 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "/?"))
    {
        puts("Simple SNDv3 player - Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
             "Usage:\n"
             "   snd <filename.snd>\n"
            );
        return 1;
    }
    
    song song = ReadSND(argv[1]);
    
    /* TODO: Handle the snd (dump it somewhere or something) */
    
    AnalyzeSND(song);
    DumpSND(song);
    
    //TickSpacer = 7 / 2.0;
    
    Patternize(song);
    PlaySND(song);
    
    return 0;
}
